#!/usr/local/bin/wish
##
## This is the second attempt at creating a
## Macintosh-like network chooser in tcl.
## The idea for this version is to make a generic
## interface that accepts plug-in for protocol-specific
## functionality
##
## Ethan Gold <etgold@cs.vassar.edu> 2/24/98
##
######## plugin API description: ########
## see the API_Description file that came
## with the distribution for a more complete
## description of the API.
## - A listbox for listing entities into
## - A frame for posting any kind of buttons
##   or other GUI elements into
## - Paths to bin directores and stuff as
##   supplied by major protocol plugins or config files
## - A current "zone" or "workgroup" variable
## - A simple interface to the relavent local selection,
##   hopefully more robust than direct X selection
## - A function to call on initialization to register
##   the plugin with the main chooser
##
#### In turn, the plugins must supply the following
#### procedures to the main chooser in the form of
#### "pluginname.procname"
## - An invocation procedure to be called when
##   the plugin's icon is clicked in the service browser
## - An "unloading" procedure to be called when another
##   service is selected
######## end plugin API description ########

######## global defines area ########
## set this to point to your library installation directory
set libdir "/usr/local/lib/tkchooser2"
set version "0.55"

set plugindir "plugins"
set icondir "icons"
set genericicon "generic.pnm"
set configfilename "chooser.cfg"
set userconfdir "$env(HOME)/.tkchooser"

## tcl default
set propfont "Helvetica -12 bold"
set fixedfont "fixed"

## set error counter
set errno 1
set debug 0

## variable containing name of listbox
## available to plugins
set pluglist ".rightside.top.pluglist"
set pluglistlabel ".rightside.top.pluglabel"
set plugframe ".rightside.bot"

## space for plugins to store state information
## accessed thru functions
set plug_globals(dummy) dum
unset plug_globals(dummy)

## kludge for plugin textentry variables
set plug_globalflag ""

set curr_pluglist_item ""
set currplug ""
set currproto ""

## set up major protocol service arrays
set appletalkservices(dummy) "dum"
unset appletalkservices(dummy)
set smbservices(dummy) "dum"
unset smbservices(dummy)
set ipxservices(dummy) "dum"
unset ipxservices(dummy)
set miscservices(dummy) "dum"
unset miscservices(dummy)
set appletalkflag 0
set smbflag 0
set ipxflag 0
set miscflag 0

######## end global defines ########

puts "using libdir $libdir"

######## get configuration options from filesystem ###########
## load global config file from the library directory
## and then load the major protocol
## handlers which will define the high-level layer-independant
## network API for each major protocol
##
if { [file exists $libdir/$configfilename] } {
    source $libdir/$configfilename
} else {
    puts "unable to find global configfile: $libdir/$configfilename."
    puts "flying by the seat of our pants - expect a crash"
}

## source's the user's options
if { [file exists $userconfdir/$configfilename] } {
    source $userconfdir/$configfilename
}

######## end configuration options from filesystem ###########


######## Build main Chooser window ########

## frame for left side
frame .leftside
## frame for right side
frame .rightside
## frame for the middle
frame .middle

########### build leftside frame
## a frame for the top left half of the window
frame .leftside.top
## a main scrolled canvas listbox
label .leftside.top.servlabel -text "Services" -border 2
canvas .leftside.top.services -relief sunken -border 2 \
	-width 175 -height 200 -background white \
	-yscrollcommand ".leftside.top.servscroll set" \
	-takefocus 0

scrollbar .leftside.top.servscroll -command ".leftside.top.services yview" \
	-takefocus 0

pack .leftside.top.servlabel -side top
pack .leftside.top.services -side left -fill both
pack .leftside.top.servscroll -side left -fill y

## a frame for the bottom left half of the window
frame .leftside.bot
## a zone selection listbox
label .leftside.bot.zonelabel -text "Zones/Workgroups:" -border 2
listbox .leftside.bot.zones -width 25 -exportselection 0 \
	-yscrollcommand ".leftside.bot.zonescroll set"
scrollbar .leftside.bot.zonescroll -command ".leftside.bot.zones yview" \
	-takefocus 0

pack .leftside.bot.zonelabel -side top
pack .leftside.bot.zones -side left -fill both
pack .leftside.bot.zonescroll -side left -fill y

## pack top and bottom halves of leftside
pack .leftside.top -side top
pack .leftside.bot -side bottom

############## build middle frame
button .middle.close -width 10 -text "Close" -command {exit}
## create the menubutton and leave the menu building for later
menubutton .middle.protocols -text "Protocols:" -relief raised
menu .middle.protocols.protomenu
label .middle.label -text " "

pack .middle.close -side bottom -anchor s
pack .middle.label -side top
pack .middle.protocols

############## build rightside frame
## a frame for the top right half of the window
frame .rightside.top

## a listbox to be made available to plugins
## as defined in the global section
label $pluglistlabel -text "Entities:" -border 2
listbox $pluglist -width 25 -height 12 -exportselection 0 \
	-yscrollcommand [list $pluglist"scroll" set]
scrollbar $pluglist"scroll" -command "$pluglist yview" \
	-takefocus 0

pack $pluglistlabel -side top
pack $pluglist -side left -fill both
pack $pluglist"scroll" -side left -fill y

## an empty frame for use by plugins
## on the bottom right section of the main window
frame $plugframe -width 200 -height 195 -relief groove -border 2

## a status widget in the plugin frame
message $plugframe.status -width 200 -text "ready."
pack $plugframe.status -side bottom -anchor s -fill x

## pack top and bottom halves of rightside
pack .rightside.top -side top -anchor n
pack $plugframe -side bottom -fill both

## pack all the frames into the main window
pack .leftside -side left -anchor n
pack .middle -side left -anchor n -fill y
pack .rightside -side right -anchor n

update
######## End Build main window ########

############# begin procedures and functions ##################

## procedure to load in plugins from plugin directory
proc load_plugins {} {
    global libdir plugindir userconfdir
    
    puts "loading global plugins..."

    set plugins [open "|/bin/ls $libdir/$plugindir" r]
    while { [gets $plugins line] != -1 } {
	if {[file readable "$libdir/$plugindir/$line"]} {
	    if {[regexp {.*\.tcl$} $line]} {
		puts "found plugin: $line"
		source "$libdir/$plugindir/$line"
	    } else { puts "found junk in plugin directory called \"$line\"" }
	}
    }
    catch {close $plugins}
    puts "finished loading global plugins."

    puts "loading local plugins..."
    if {[file readable $userconfdir/$plugindir]} {
	set plugins [open "|/bin/ls $userconfdir/$plugindir" r]
	while { [gets $plugins line] != -1 } {
	    if {[file readable "$userconfdir/$plugindir/$line"]} {
		puts "found plugin: $line"
		source "$userconfdir/$plugindir/$line"
	    }
	}
	catch {close $plugins}
    }
    puts "finished loading local plugins"
}

################ Plugin API ######################

## functions to return the names of GUI
## elements that the plugins are allowed to manipulate
proc plug_frame {} {
    global plugframe
    return $plugframe
}
proc plug_list {} {
    global pluglist
    return $pluglist
}
proc plug_list_label {} {
    global pluglistlabel
    return $pluglistlabel
}

proc get_curr_item {} {
    global curr_pluglist_item
    return $curr_pluglist_item
}

proc get_glob {index} {
    global plug_globals
    if {[info exists plug_globals($index)]} {
	return $plug_globals($index)
    } else {
	return ""
    }
}

proc set_glob {index value} {
    global plug_globals
    set plug_globals($index) $value
}

proc get_propfont {} {
    global propfont
    return $propfont
}

proc get_fixedfont {} {
    global fixedfont
    return $fixedfont
}

proc debug {} {
    global debug
    return $debug
}

## display an error message
proc error {caller errormessage} {
    global errno
    toplevel .error_$errno -class Dialog
    label .error_$errno.label -relief groove -text "$caller error:"
    message .error_$errno.message -width 300 -text "$errormessage"
    button .error_$errno.ok -text "Ok" -command [list destroy .error_$errno]
    pack .error_$errno.label -fill x
    pack .error_$errno.message -fill x
    pack .error_$errno.ok
    wm title . "tkchooser error #$errno"
    incr errno
}

## function plugin calls to register itself
proc register_plugin {major plugname plugpubname} {
    global appletalkservices smbservices ipxservices miscservices
    global appletalkflag smbflag ipxflag miscflag
    upvar $major protocol

    if {[debug]} {
	puts "register_plugin called with $major $plugname $plugpubname"
    }

    ## this is uglee
    if {[string compare $major "appletalk"] == 0} {
	## if the major protocol is not running
	## then just return
	if {!$appletalkflag} {
	    puts "appletalk services not available"
	    return
	}

	## otherwise, register the plugin
	if {[debug]} {
	    puts "registering plugin: $plugname, $plugpubname"
	}
	set appletalkservices($plugname) $plugpubname	
    } elseif {[string compare $major "smb"] == 0} {
	## if the major protocol is not running
	## then just return
	if {!$smbflag} {
	    puts "smb services not available"
	    return
	}

	## otherwise, register the plugin
	if {[debug]} {
	    puts "registering plugin: $plugname, $plugpubname"
	}
	set smbservices($plugname) $plugpubname	
	
    } elseif {[string compare $major "ipx"] == 0} {
	## if the major protocol is not running
	## then just return
	if {!$ipxflag} {
	    puts "ipx services not available"
	    return
	}

	## otherwise, register the plugin
	if {[debug]} {
	    puts "registering plugin: $plugname, $plugpubname"
	}
	set ipxservices($plugname) $plugpubname	

    } elseif {[string compare $major "misc"] == 0} {
	## if the major protocol is not running
	## then just return
	if {!$miscflag} {
	    puts "misc services not available"
	    return
	}

	## otherwise, register the plugin
	if {[debug]} {
	    puts "registering plugin: $plugname, $plugpubname"
	}
	set miscservices($plugname) $plugpubname		

    } else {
	puts "unsuppored protocol: $major"
    }
    
    ## all other function names can be deduced
    ## from the plugin name

}

################ End Plugin API ######################

## internal function to build services canvas
proc build_services {major} {
    global $major libdir icondir genericicon userconfdir
    upvar $major protocol
    
    if {[debug]} {
	puts "building $major services: [array names protocol]..."
    }

    .leftside.top.services delete all

    ## foreach plugin defined for this protocol
    ## create it's icon in the canvas and bind
    ## single-click to it's startup function

    set spacing 45
    set rownum 1
    set leftside 1
    foreach plugin [array names protocol] {
	set icon [$plugin.geticon]
	set iconpathglob "$libdir/$icondir"
	set iconpathloc "$userconfdir/$icondir"
	if {[debug]} {
	    puts "activating $protocol($plugin) with icon $iconpathglob/$icon"
	}
	
	if {!$leftside} { 
	    set X [expr $spacing * 3]
	    set leftside 1
	} else {
	    set X $spacing
	    set leftside 0
	}
	
	if {$rownum == 1} {
	    set Y [expr $spacing/1.5]
	} else {
	    set Y [expr $spacing * $rownum]
	}

	## put in a check for the icon and make a generic one
	if {[file readable $iconpathloc/$icon]} {
	    image create photo $plugin -file $iconpathloc/$icon
	} elseif {[file readable $iconpathglob/$icon]} {
	    image create photo $plugin -file $iconpathglob/$icon
	} else {
	    puts "tkchooser2: could not find or read $icon for $plugin plugin"
	    image create photo $plugin -file $iconpathglob/$genericicon
	}

	.leftside.top.services create image \
		$X $Y -anchor center -image "$plugin" -tag "$plugin"
	
	.leftside.top.services create text \
		[expr $X] [expr $Y + $spacing/2] \
		-text [$plugin.getpubname] \
		-font fixed -fill black \
		-tag "$plugin"

	.leftside.top.services bind $plugin <ButtonPress> \
		[list startplug $plugin]

	## make sure the scroll region for the canvas will let
	## us see all the icons
	if {$Y >= 200} {
	    .leftside.top.services configure \
		    -scrollregion "0 0 300 [expr $Y + $Y/2]"
	}

	if {$leftside} {set rownum [expr $rownum + 1]}
    }
    update
    if {[debug]} {
	puts "finished building $major services"
    }
}

## procedure to remove services
proc remove_services {} {
    .leftside.top.services delete all
}

## procedure to startup plugins when their
## icons are clicked
proc startplug {plugin} {
    global currplug curr_pluglist_item plug_globals plug_globalflag

    ## if there's something running, then shut it off
    ## and clear the plugin globals
    if {[string compare $currplug ""] != 0} {
	$currplug.stop
	foreach index [array names plug_globals] {
	    unset plug_globals($index)
	}
	set plug_globalflag ""
    }
    
    ## set up and start the new plugin
    set currplug $plugin
    set curr_pluglist_item ""
    $currplug.start
}

## procedure called from doubleclick on
## plugin list window to call the current
## plugin's doubleclick function
proc doubleclick_currplug {} {
    global currplug
    $currplug.doubleclick
}

## procedure to start up a new protocol
proc startproto {protocol} {
    global currproto menuproto currplug pluglistlabel plugframe
    if {[debug]} {
	puts "starting $protocol module"
    }
    $pluglistlabel configure -text "Entities:"

    ## stop any running plugins
    if {[string compare $currplug ""] != 0 } {
	$currplug.stop
    }
    ## stop the running protocol
    if {[string compare $currproto ""] != 0 } {
	$currproto.stop
    }
    
    ## reset and start the new protocol
    set currproto $protocol
    set menuproto $protocol
    $plugframe.status configure -text "starting $protocol"
    update
    $protocol.start
    .leftside.top.servlabel configure -text "$protocol Services:"
    $plugframe.status configure -text "ready."
    update
}


######### End Procedures and Functions ############

######### Resume startup execution ############

## now that we have the config file loaded, the main
## window built, and the functions defined, check for
## protocol module files.

## check for appletalk module
if { [file exists $libdir/appletalk.tcl] && $appletalkflag } {
    puts "checking for Appletalk module"
    source $libdir/appletalk.tcl
}

## check for SMB module
if { [file exists $libdir/smb.tcl] && $smbflag } {
    puts "checking for SMB module"
    source $libdir/smb.tcl
}

## check for IPX module
if { [file exists $libdir/ipx.tcl] && $ipxflag } {
    puts "checking for IPX module"
    source $libdir/ipx.tcl
}

## check for Misc module
if { [file exists $libdir/misc.tcl] && $miscflag } {
    puts "checking for Misc module"
    source $libdir/misc.tcl
}

## source the fileselction box code so everyone can use it
source $libdir/fileselect.tcl

## search the plugin directory and register 'em
load_plugins

## build a protocol selection menu based on what protocols
## are defined as active

if {$appletalkflag} {
    .middle.protocols.protomenu add radiobutton \
	    -label "Appletalk" -variable menuproto -value "appletalk" \
	    -command "build_services appletalkservices; startproto appletalk"
    #-command "startproto appletalk; build_services appletalkservices"
}
if {$smbflag} {
    .middle.protocols.protomenu add radiobutton \
	    -label "SMB" -variable menuproto -value "smb" \
	    -command "build_services smbservices; startproto smb"
    #-command "startproto smb; build_services smbservices"
}
if {$ipxflag} {
    .middle.protocols.protomenu add radiobutton \
	    -label "IPX" -variable menuproto -value "ipx" \
	    -command "build_services ipxservices; startproto ipx"
    #-command "startproto smb; build_services ipxservices"
}
if {$miscflag} {
    .middle.protocols.protomenu add radiobutton \
	    -label "Misc" -variable menuproto -value "misc" \
	    -command "build_services miscservices; startproto misc"
    #-command "startproto misc; build_services miscservices"
}
.middle.protocols configure -menu .middle.protocols.protomenu


## select and activate the default protocol
set service "services"
set service $defaultprotocol$service
if {[debug]} {
    puts "debug: trying to build $service"
}
build_services $service
startproto $defaultprotocol

wm title . "Tkchooser $version"

############## Event Bindings ################
## plugin listbox
bind $pluglist <ButtonRelease> {
    if {[debug]} {
	puts "[$pluglist curselection]"
    }
    if {[string compare [$pluglist curselection] ""] != 0} {
	set curr_pluglist_item [$pluglist get [$pluglist curselection]]
    }
}

## module zone/workgroup listbox
bind .leftside.bot.zones <ButtonRelease> {
    #$currproto.setcurrzone [selection get]
    set newzone  [.leftside.bot.zones get [.leftside.bot.zones curselection]]
    $currproto.setcurrzone "$newzone"
    if {[string compare $currplug ""] != 0} {
	$currplug.newzone
    }
}

bind .leftside.bot.zones <Return> {
    #$currproto.setcurrzone [selection get]
    set newzone  [.leftside.bot.zones get [.leftside.bot.zones curselection]]
    $currproto.setcurrzone "$newzone"
    if {[string compare $currplug ""] != 0} {
	$currplug.newzone
    }
}



bind $pluglist <Double-1> {
    #set curr_pluglist_item [selection get]
    if {[string compare [$pluglist curselection] ""] != 0} {
	set curr_pluglist_item [$pluglist get [$pluglist curselection]]
    }
    doubleclick_currplug
}

bind $pluglist <Return> {
    if {[string compare [$pluglist curselection] ""] != 0} {
	set curr_pluglist_item [$pluglist get [$pluglist curselection]]
    }
    doubleclick_currplug    
}

bind .middle.close <Return> {
    .middle.close invoke
}

############# End Event Bindings ###############

